@@ -30,7 +30,7 @@ use std::{
 
 use anyhow::Error as AnyError;
 use futures::{future::BoxFuture, stream::TryStreamExt};
-use rusoto_core::RusotoError;
+use rusoto_core::{request::BufferedHttpResponse, RusotoError};
 use rusoto_s3::{GetObjectError, GetObjectRequest, HeadObjectError, HeadObjectRequest, S3};
 
 pub use rusoto_core::Region;
@@ -238,7 +238,7 @@ impl<C: S3 + Clone + Send + Sync + 'static> CondowClient for S3ClientWrapper<C>
 
             let bytes_hint = response
                 .content_length
-                .map(|s| BytesHint::new_exact(s as usize))
+                .map(|s| BytesHint::new_exact(s as u64))
                 .unwrap_or_else(BytesHint::new_no_hint);
 
             let stream = if let Some(stream) = response.body {
@@ -258,7 +258,12 @@ impl<C: S3 + Clone + Send + Sync + 'static> CondowClient for S3ClientWrapper<C>
 
 fn get_obj_err_to_download_err(err: RusotoError<GetObjectError>) -> CondowError {
     match err {
-        RusotoError::Service(err) => CondowError::new_remote("remote error").with_source(err),
+        RusotoError::Service(err) => match err {
+            GetObjectError::NoSuchKey(s) => CondowError::new_not_found(s),
+            GetObjectError::InvalidObjectState(s) => {
+                CondowError::new_other(format!("invalid object state: {}", s))
+            }
+        },
         RusotoError::Validation(cause) => {
             CondowError::new_other(format!("validation error: {}", cause))
         }
@@ -269,14 +274,18 @@ fn get_obj_err_to_download_err(err: RusotoError<GetObjectError>) -> CondowError
             CondowError::new_other("http dispatch error").with_source(dispatch_error)
         }
         RusotoError::ParseError(cause) => CondowError::new_other(format!("parse error: {}", cause)),
-        RusotoError::Unknown(_cause) => CondowError::new_other("unknown"), //.with_source(cause),
-        RusotoError::Blocking => CondowError::new_other("failed to run blocking future"),
+        RusotoError::Unknown(response) => response_to_condow_err(response),
+        RusotoError::Blocking => {
+            CondowError::new_other("failed to run blocking future within rusoto")
+        }
     }
 }
 
 fn head_obj_err_to_get_size_err(err: RusotoError<HeadObjectError>) -> CondowError {
     match err {
-        RusotoError::Service(err) => CondowError::new_remote("remote error").with_source(err),
+        RusotoError::Service(err) => match err {
+            HeadObjectError::NoSuchKey(s) => CondowError::new_not_found(s),
+        },
         RusotoError::Validation(cause) => {
             CondowError::new_other(format!("validation error: {}", cause))
         }
@@ -287,7 +296,31 @@ fn head_obj_err_to_get_size_err(err: RusotoError<HeadObjectError>) -> CondowErro
             CondowError::new_other("http dispatch error").with_source(dispatch_error)
         }
         RusotoError::ParseError(cause) => CondowError::new_other(format!("parse error: {}", cause)),
-        RusotoError::Unknown(_cause) => CondowError::new_other("unknown"), //.with_source(cause),
-        RusotoError::Blocking => CondowError::new_other("failed to run blocking future"),
+        RusotoError::Unknown(response) => response_to_condow_err(response),
+        RusotoError::Blocking => {
+            CondowError::new_other("failed to run blocking future within rusoto")
+        }
+    }
+}
+
+fn response_to_condow_err(response: BufferedHttpResponse) -> CondowError {
+    let message = if let Ok(body_str) = std::str::from_utf8(response.body.as_ref()) {
+        body_str
+    } else {
+        "<<< response body received from AWS not UTF-8 >>>"
+    };
+
+    let status = response.status;
+    let message = format!("{} - {}", status, message);
+    match status.as_u16() {
+        404 => CondowError::new_not_found(message),
+        401 | 403 => CondowError::new_access_denied(message),
+        _ => {
+            if status.is_server_error() {
+                CondowError::new_remote(message)
+            } else {
+                CondowError::new_other(message)
+            }
+        }
     }
 }
